/*
 *  Copyright 2017 The WebRTC project authors. All Rights Reserved.
 *
 *  Use of this source code is governed by a BSD-style license
 *  that can be found in the LICENSE file in the root of the source
 *  tree. An additional intellectual property rights grant can be found
 *  in the file PATENTS.  All contributing project authors may
 *  be found in the AUTHORS file in the root of the source tree.
 */

package tee3.webrtc;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.annotation.Nullable;

public class SoftwareVideoEncoderFactory implements VideoEncoderFactory {
  @Nullable
  @Override
  public VideoEncoder createEncoder(VideoCodecInfo info) {
    if (info.name.equalsIgnoreCase("VP8")) {
      return new VP8Encoder();
    }
    if (info.name.equalsIgnoreCase("VP9") && VP9Encoder.nativeIsSupported()) {
      return new VP9Encoder();
    }
    if (info.name.equalsIgnoreCase("H264") && H264Encoder.nativeIsSupported()) {
      return new H264Encoder();
    }
    if (info.name.equalsIgnoreCase("H265") && H265Encoder.nativeIsSupported()) {
      return new H265Encoder();
    }

    return null;
  }

  @Override
  public VideoCodecInfo[] getSupportedCodecs() {
    return supportedCodecs();
  }

  static VideoCodecInfo[] supportedCodecs() {
    List<VideoCodecInfo> codecs = new ArrayList<VideoCodecInfo>();

    codecs.add(new VideoCodecInfo("VP8", new HashMap<>()));
    if (VP9Encoder.nativeIsSupported()) {
      codecs.add(new VideoCodecInfo("VP9", new HashMap<>()));
    }
    if (H264Encoder.nativeIsSupported()) {
      codecs.add(new VideoCodecInfo("H264", new HashMap<>()));
    }
    if (H265Encoder.nativeIsSupported()) {
      codecs.add(new VideoCodecInfo("H265", new HashMap<>()));
    }
    return codecs.toArray(new VideoCodecInfo[codecs.size()]);
  }
}
